/**
* creates a new zoom behavior
*/
var zoom = d3.zoom().on("zoom", zoomHandler);

/**
* creates svg object and associated attributes
* applies the zoom behavior to svg
*/
var svg = d3.select("svg")
    .call(zoom),
width = svg.attr("width"),
height = svg.attr("height");

/**
* creates node object and associated attributes
*/
var node,
r=10,
color = d3.scaleOrdinal()
    .domain(["citing", "input", "cited"])
    .range([' #01d7c0', ' #8b90fe ', '  #a15eb2 ']),
yscale = d3.scaleOrdinal()
    .domain(["citing", "input", "cited"])
    .range([0, 200, 400]),
toRemove;

/**
* creates link object
*/
var link;

/**
* creates a background
* creates a click functionality of the background
*/
var rect = svg.append("rect")
    .attr("x", 0)
    .attr("y", 0)
    .attr("height", height)
    .attr("width", width)
    .style("fill", 'white')
    .on('click', clickRect);

/**
* creates a new simulation
* updates the positions of the links and nodes when the 
  state of the layout has changed (simulation has advanced by a tick)
*/
var simulation = d3.forceSimulation()
    .force("link", d3.forceLink().id(function(d) {return d.doi;}).distance(100).strength(1))
    .force("collide", d3.forceCollide(50))
    .force("charge", d3.forceManyBody().strength(-30))
    .force("center", d3.forceCenter(width/2, height/2))
    .force("yscale", d3.forceY().strength(1).y(function(d) {return yscale(d.group)}));

/**
* creates group element
*/
var g = svg.append("g")
    .attr("class", "everything")

/**
* loads JSON data and calls the update function
*/
d3.json("json_text.json").then(function(graph) {
    update(graph.links, graph.nodes);
})

/**
* calls update functions for links and nodes
* adds the nodes and links to the simulation
* @param {object} nodes - nodes
* @param {object} links - links
*/
function update(links, nodes) {
    updateLinks(links);
    updateNodes(nodes);

    simulation
        .nodes(nodes)
        .on("tick", tickHandler);
    simulation.force("link")
        .links(links);
    
    d3.selectAll(".link").attr('marker-end', function(d) {return updateMarker("#999", d.target);});
}

/**
* initializes and shows links
* @param {object} links - links
*/
function updateLinks(links) {
    link = g.append("g")
        .selectAll(".link")
        .data(links)
        .enter()
        .append("line")
        .style("stroke-width", "1px")
        .style("stroke", "#999")
        .attr("class", "link");
}

/**
* initializes and shows nodes with circles and texts
* creates a new drag behavior and applies it to the circles
* creates a click functionality of the circles and texts
* @param {object} nodes - nodes
*/
function updateNodes(nodes) {
    node = g.selectAll(".node")
        .data(nodes)
        .enter()
        .append("g")
        .attr("class", "node")
        .attr("initial_x", function(d) {return d.dx;})
        .attr("initial_y", function(d) {return d.dy;})
        .call(d3.drag()
            .on("start", dragstarted)
            .on("drag", dragged)
        );

    node.append("circle")
        .attr("class", "circle")
        .attr("r", function(d) {return r+d.citations*0.1})
        .style("fill", function(d){ return color(d.group)})
        .on('click', clickNode);

    node.append("text")
        .attr("class", "text") 
        .style("font-size", "15px")
        .style('pointer-events', 'auto')
        .text(function (d) {return firstauthor(d.author);})
        .on('click', clickNode);
}

/**
* creates arrowhead and returns its url
* @param {string} color - color of arrowhead
* @param {string} target - target-node
*/
function updateMarker(color, target) {
    var radius = r+target.citations*0.1;
    svg.append('defs').append('marker')//arrowhead
        .attr('id',color.replace("#", "")+radius)
        .attr('viewBox','-0 -5 10 10')
        .attr('refX',radius+10)
        .attr('refY',0)
        .attr('orient','auto')
        .attr('markerWidth',10)
        .attr('markerHeight',15)
        .attr('xoverflow','visible')
        .append('svg:path')
        .attr('d', 'M 0,-5 L 10 ,0 L 0,5')
        .attr('fill', color)//arrowhead color
        .style('stroke','none');
    return "url(" + color + radius + ")";
};

/**
* colors the circle and its links black and removes the previous markings
* @param {object} node - node
*/
function clickNode(node) {
    if(toRemove){
        d3.select(toRemove).selectAll(".circle").style("stroke","none")
    }
    toRemove = this.parentNode;
    d3.select(this.parentNode).selectAll(".circle").style("stroke","black")
    marklink(node)
    textfunc(node)
}

/**
* removes the markings of the circles and their links
*/
function clickRect() {
    d3.selectAll(".circle").style("stroke", "none")
    d3.selectAll(".link")
        .style("stroke", "#999")
        .attr('marker-end', function(d) {return updateMarker('#999', d.target);})
    document.getElementById('textbox').innerHTML = "Click node";
}

/**
* sets color of link (line and arrowhead) to black if it is directly connected to node
* and to grey otherwise
* @param {object} node - node
*/
function marklink(node){
    d3.selectAll(".link")
        .style("stroke", function(o) {
            return isLinkForNode(node, o) ? "black" : "#999";})
        .attr('marker-end', function(o) {
            return isLinkForNode(node, o) ? updateMarker('#000000', o.target) : updateMarker('#999', o.target);})
}

/**
* returns true if link is directly connected to node and false if it is not
* @param {object} node - node
* @param {object} link - link
*/
function isLinkForNode(node, link){
    return link.source.index == node.index || link.target.index == node.index;
}

/**
* returns last name of first author
* @param {string} authors - the comma-separated string of authors
*/
function firstauthor(authors){
    if (/,/.test(authors)==false){
        var firstauthor=/^.*\s+([\w\-]+)[\.\s]*$/.exec(authors)
    }
    else {
        var firstauthor=/^[\s\w\.\-]*\s([\w\-]+)[\.\s]*,.*$/.exec(authors)
    }
    return firstauthor[1]
}

/**
* outputs node info to textbox
* @param {object} node - data of current node
*/
function textfunc(node){
    document.getElementById('textbox').innerHTML = "Title:" + '</br>' + node.name +
    '</br>' +'</br>'+"Author:"+ '</br>' +node.author+'</br>'+'</br>'+"Year:"+'</br>'
    +node.year+'</br>'+'</br>'+"doi:"+'</br>'+node.doi;
}

/**
* updates the positions of the links and nodes
*/
function tickHandler() {
    link.attr("x1", function (d) {return d.source.x;})
        .attr("y1", function (d) {return d.source.y;})
        .attr("x2", function (d) {return d.target.x;})
        .attr("y2", function (d) {return d.target.y;});
    node.attr("transform", function (d) {return "translate(" + d.x + ", " + d.y + ")";});
}

/**
* initializes the dragging of the node
* @param {object} node - data of current node
*/
function dragstarted(node) {
    if (!d3.event.active) 
        simulation.alphaTarget(0.3).restart()
    node.fx = node.x;
    node.fy = node.y;
}

/**
* applies the dragging to the node
* @param {object} node - data of current node
*/
function dragged(node) {
    node.fx = d3.event.x;
    node.fy = d3.event.y;
}

/**
* resets the positions of the nodes
*/
function resetGraph() {
    d3.selectAll(".node").each(function(d) {
        d.fx = d.initial_x;
        d.fy = d.initial_y;
    })
}

/**
* applies the transformation (zooming or dragging) to the g element
*/
function zoomHandler() {
    d3.select('g').attr("transform", d3.event.transform);
}

/**
* transforms svg so that that the zoom is reset
*/
function resetZoom() {
    d3.select('svg')
        .call(zoom.scaleTo, 1);
}

/**
* transforms svg so that it is centered
*/
function center() {
    d3.select('svg')
        .call(zoom.translateTo, 0.5 * width, 0.5 * height);
}